describe('waiter.js suite', waiter.mixin(function(args, _store, waiter){

  describe('#load called first', function(){

    before(function(){
      waiter.load('foo', 'bar');
      waiter.load('Foo', 'bar');
    });

    after(function(){
      waiter.unload('foo');
      waiter.unload('Foo');
    });

    it('adds key to loadStack', function(){
      assert.property(_store.loadStack, 'foo', 'key not added to loadStack');
    });

    it('creates deferred when no deferred present', function(){
      assert.property(_store.loadStack.foo, 'resolve', 'fails to create deferred');
    });

    it('immediately resolves deferred when no deferred present', function(){
      assert.equal(_store.loadStack.foo.state(), 'resolved', 'fails to resolve deferred');
    });

    it('resolves deferred with passed in value', function(){
      _store.loadStack.foo.done(function(val){
        assert.equal(val.val, 'bar', 'fails to resolve with correct value');
      });
    });

    it('adds key to global namespace if it starts with lowercase', function(){
      assert.property(waiter, 'foo', 'does not add to global namespace');
    });

    it('does not add key to global namespace if it start with lowercase', function(){
      assert.notProperty(waiter, 'Foo', 'adds key to global namespace');
    });

  });

  describe('#load called second', function(){

    before(function(){
      _store.loadStack.foo = new $.Deferred();
      sandbox.spy($, 'Deferred');
      sandbox.spy(_store.loadStack.foo, 'resolve');
      waiter.load('foo', 'bar');
    });

    after(function(){
      sandbox.restore();
      waiter.unload('foo');
    });

    it('resolves pending deferred', function(){
      assert.isFalse($.Deferred.calledWithNew(), 'creates a new deferred');
      assert(_store.loadStack.foo.resolve.calledOnce, 'fails to invoke resolve method on deferred');
    });
  });

  describe('#waitFor', function(){

    // TODO: probably should break this up a bit, maybe separating out error
    // check tests
    before(function(){
      sandbox.spy(waiter, 'watchForDependencyTimeout');
      sandbox.stub(window, 'setTimeout');
      sandbox.spy($, 'when');
      sandbox.stub(waiter, 'waitForRemote');
      sandbox.stub(console, 'log');
      waiter.waitFor('foo');
    });

    after(function(){
      sandbox.restore();
      waiter.unload('foo');
    });

    it('adds key to loadStack', function(){
      assert.property(_store.loadStack, 'foo', 'fails to add key');
    });

    it('creates deferred if not already present', function(){
      assert.property(_store.loadStack.foo, 'resolve', 'fails to create deferred');
    });

    it('creates unresolved deferred if not already present', function(){
      assert.equal(_store.loadStack.foo.state(), 'pending', 'fails to create pending deferred');
    });

    it('queues a timeout error check with correct args', function(){
      assert(waiter.watchForDependencyTimeout.calledWith(_store.loadStack['foo'], 'foo'), 'waiter.watchForDependencyTimeout not called or called with wrong args');
    });

    it('calls $.when with deferred', function(){
      assert($.when.calledWith(_store.loadStack.foo), '$.when not called with deferred');
    });

    it('rejects deferred on timeout', function(){
      setTimeout.callArg(0);
      assert.equal(_store.loadStack.foo.state(), 'rejected', 'fails to reject deferred');
    });

    it('delegates to #waitForRemote when detecting a remote resource', function(){
      waiter.waitFor('Remote');
      waiter.load('Remote', 'http://foo.com');
      assert(waiter.waitForRemote.calledOnce, 'called #waitForRemote ' + waiter.waitForRemote.callCount + ' times instead of once');
      waiter.unload('Remote');
    });

  });

  describe('#waitForRemote', function(){
    before(function(){
      sandbox.stub($, 'ajax').returns({done: sinon.stub().returns({fail: sinon.stub()}), then: sinon.stub()});
    });

    afterEach(function(){
      waiter.unload('Remote');
      waiter.unload('Foo');
      $.ajax.reset();
    });

    after(function(){
      sandbox.restore();
    });

    it('loads url when config module specifies nothing else', function(){
      waiter.load('Foo', 'http://foo.com');
      waiter.waitFor('Foo');
      assert($.ajax.calledWithMatch({url: 'http://foo.com'}), '$.ajax fails to be called with right url');
    });

    it('does not call $.ajax when called once already', function(){
      waiter.load('Remote', 'http://foo.com');
      waiter.waitFor('Remote');
      waiter.waitFor('Remote');
      assert.isTrue($.ajax.calledOnce, 'repeatedly calls $.ajax');
    });

    it('passes options properly to $.ajax', function() {
      var options = {params: {'bar': 'baz'}};
      waiter.load('Remote', 'http://foo.com', options);
      waiter.waitFor('Remote');
      assert($.ajax.calledWithMatch({data: options.params}), 'fails to pass in params');
    });

    it('exports specified object', function() {
      var config = {exports: 'Foo'};
      $.ajax.returns($.Deferred().resolve({Foo: 'bar'}));

      waiter.load('Remote', 'http://foo.com', config);
      waiter.waitFor('Remote').then(function(Foo){
        assert.equal(Foo, 'bar', 'fails to return exported object');
      });
    });
  });

  describe('#unload', function(){
    before(function(){
      waiter.load('foo', 'bar');
    });

    it('deletes property altogether', function(){
      waiter.unload('foo');
      assert.isNull(_store.loadStack.foo, 'fails to delete property');
      assert.isNull(_store.remoteStack.foo, 'fails to delete property');
      assert.isNull(waiter.foo, 'fails to delete it from global namespace if lower case');
    });
  });

}));
