/**
 * waiter.js
 *
 */
;(function(window, document, $){
  "use strict";

  var settings = {
    // default timeout for local modules, in ms
    timeout: 15000,
    // store custom timeouts in format 'module': timeout in ms
    moduleTimeouts: {},
    // for remotes, each value represents one try, to run after the specified ms since previous try
    retries: [5000, 8000, 5000],
    // custom progress callback for slow remotes. this will be called once for
    // every try after the first try in the 'retries' array above.
    progressHandler: null,
    // custom error handler
    errorHandler: null,
    // log errors to console
    debug: true 
  },
  // private store
  _store = {
    // storage for deferreds
    loadStack: {},
    remoteStack: {},
    constructors: {}
  },

  waiter = {

    /**
     * Request a module. If it is not already in the system, create
     * a deferred for it, to be resolved when it actually loads.
     *
     * @method waitFor
     * @param {String|Array} args* One or more strings, or an array of
     * strings, representing module keys
     * @returns {Promise} A promise that will be resolved with the requested modules in passed order
     * @api public
     */
    waitFor: function(){
      var args = Array.prototype.slice.call(arguments),
          deferredArray = [],
          self = this,
          key, dfd;

      // allow for passing in an array rather than individual arguments
      if(args && args.length && (args[0] instanceof Array))
        args = args[0];

      for(var i = 0, len = args.length; i < len; i++){
        key = args[i];
        dfd = _store.loadStack[key];

        if(!dfd){
          // if this key doesn't already exist in the loadStack, add it,
          // expecting the #load method to resolve it.
          dfd = _store.loadStack[key] = new $.Deferred();
          waiter.watchForDependencyTimeout(dfd, key);
        }

        deferredArray.push(dfd);
      }

      // Return the deferreds to the requester, checking for and fetching remotes in
      // a `then` pipe.
      return $.when.apply($, deferredArray).then(function(){
        var args = Array.prototype.slice.call(arguments),
            item;

        for (var i = 0, len = args.length; i < len; i++) {
          item = args[i];
          args[i] = waiter.isRemote(item.val) ?
             _store.remoteStack[item.key] || waiter.waitForRemote(item) :
             item.val || item;
        }

        // Provide common globals to module callbacks after dependencies,
        // allowing a small compression optimization of local variables.
        args = args.concat([self, window, document]);

        return $.when.apply($, args);
      });
    },

    /**
     * Provides a method interface for configurable options to maintain some semblance of privacy
     *
     * @method config
     * @param  {Object} envelope
     * @return {void}
     * @api public
     */
    config: function(envelope) {
      $.extend(settings, envelope);
      _store.errorHandler = settings.errorHandler || this.errorHandler;
      _store.progressHandler = settings.progressHandler || this.progressHandler;
    },

    /**
     * First point of entry for modules to register with the system.
     *
     * ### Examples
     * 
     *     // any type can be loaded
     *     waiter.load("SomeKey", {data: [1,2,3,4]});
     *     
     *     // if it's a url, we'll try to load it
     *     waiter.load('myData', 'http://my.dynamic-remote.com/');
     *     
     * 
     * @method load
     * @see  register
     * @param {String} key The module name/key.
     * @param {Varied} val A module that should be used to resolve those waiting for `key` or a url that should be fetched.
     * @param {Object} [options] Optional specifying handling of module.
     * @param {String} [options.exports] Specifies where the module is in the namespace e.g. `'_'` for underscore
     * @param {String} [options.dataType] dataType to be passed to jQuery ajax call
     * @api public
     */
    load: function(key, val, options, namespace){
      /**
       * First point of entry for modules to register with the system.
       *
       * @method load
       * @param {String} key The module name/key.
       * @param {Varied} val The module value.
       */
      var k;

      if(typeof key == 'string') {
        waiter.register.call(this, val, key, options, namespace);
      }
      else {
        for(k in key) {
          if(key.hasOwnProperty(k))
            waiter.register.call(this, k, key[k], options, namespace);
        }
      }
      
    },

    /**
     * Register with the dependency management system, resolving any existing
     * deferreds with the same key (because a module already waited for it
     * before it could register), or creating a resolved defered otherwise.
     *
     * @method register
     * @param {String} key The module name/key.
     * @param {Varied} val The module value.
     * @api private
     */
    register: function(val, key, options, namespace){
      var loadKey = _store.loadStack[key],
          envelope = {
            key: key,
            val: val,
            options: options
          };

      // check to see if there is already a key for this module in the
      // loadStack. If so, and if it's a deffered, and if it is still
      // pending, resolve it.
      if(loadKey && loadKey.state && loadKey.state() == 'pending') {
        loadKey.resolve(envelope);
      }
      else if(!loadKey){
        _store.loadStack[key] = $.Deferred().resolve(envelope);
      }

      if (waiter.isEvent(key)) {
        _store.loadStack[key] = new $.Deferred();
      }

      // lowercase objects are assumed to be instances or literals, not in need of
      // further instantiation, and thus are attached to the namespace for
      // global usage
      if(/^[a-z]/.test(key)) {
        this[key] = envelope.val;
      }
      else if (envelope.val) {
        _store.constructors[waiter.getModuleName(key)] = envelope.val;
      }

      if(namespace)
        namespace[waiter.getModuleName(key)] = envelope;
    },

    /**
     * Checks a key to see if it fits our event syntax
     *
     * @method  isEvent
     * @param  {String}  key
     * @return {Boolean}
     * @api private
     */
    isEvent: function(key) {
      return key.indexOf("#") === 0 || key.indexOf("Event:") === 0;
    },

    /**
     * Checks a val to see if it is a url
     *
     * @method  isRemote
     * @param  {String}  val
     * @return {Boolean}
     * @api private
     */
    isRemote: function(val) {
      return typeof val == 'string' && ~val.indexOf('/');
    },

    /**
     * Special waitFor method that deals with remote modules.
     *
     * @method waitForRemote
     * @param {String} key The module key that should be prepended with
     * 'Remote:' to access the envelope.
     * @api private
     */
    waitForRemote: function(envelope){
      
      var dfd = _store.remoteStack[envelope.key] = new $.Deferred(),
          options = $.extend(envelope.options || {}, {key: envelope.key, url: envelope.val});
      
      // if we received nothing but a url, assume it's a script and do
      // nothing but get it
      // TODO: allow for non-script requests (for example, to request json
      // data)
      if (options.dataType === "css" || /\.css(\#|\?|$)/.test(options.url)) {
        options.dataType = "text";
        options.isStylesheet = true;
      }

      if(typeof envelope.val == 'string')
        waiter.fetch(options, dfd);

      return dfd.progress(_store.progressHandler);
    },

    /**
     * Recursive method that retries an AJAX request with gaps specified in settings.retries. The passed
     * Deferred is resolved when the request is completed. Calls settings.errorHandler on a failure.
     *
     * @method  fetch
     * @param  {Object} options Used to population $.ajax request
     * @param  {Deferred} dfd To be resolved on completion of request
     * @param  {Number} Depth counter for recursion
     * @return {void}
     * @api private
     */
    fetch: function(options, dfd, index) {
      var index = ++index || 0,
          timeout = settings.retries[index];

      if(timeout){

        // if this is not the first try, notify the optional progress callback
        // with the index of this try
        if(index)
          dfd.notify(index, options.key);

        $.ajax({
          url: options.url,
          dataType: options.dataType || 'script',
          data: options.params || {},
          crossDomain: true,
          timeout: timeout
        }).done(function(data){
          if (options.isStylesheet) {
            data = $("<link rel='stylesheet' type='text/css'/>").attr("href", options.url).appendTo("head");
          }

          if(options.callback){
            callback(data, dfd);
          }
          else if(dfd.state() == 'pending'){
            dfd.resolve(
              options.exports ? 
              (data && data[options.exports] ? data[options.exports] : window[options.exports]) : 
              data
            );
          }
        }).fail(function(data, status, jqxhr){
          if(status == 'timeout'){
            waiter.fetch(options, dfd, index);
          }
          else if(dfd.state() == 'pending'){
            _store.errorHandler(dfd, {
              data: jqxhr,
              msg: 'Failed to load module "' + options.key + '": ' + jqxhr.statusText + ' (' + jqxhr.status + ')'
            });
          }
        });
      }
      else{
        _store.errorHandler(dfd, {
          msg: 'Module "' + options.key + '" failed to load after ' + settings.retries.length + ' retries'
        });
      }
    },

    /**
     * Load dependencies, specified as an array of dependency names, for a
     * given `context` object onto that object.
     *
     * @method preload
     * @param {Object} context The object that will have its dependencies
     * preloaded.
     * @api public
     */
    preload: function(context){
      var identifiers = context.getDependencies ?
            context.getDependencies() :
            context.dependencies,
          self = this;

      if(identifiers){
        return waiter.waitFor.apply(this, identifiers).then(function(){
          var dependencies = Array.prototype.slice.call(arguments),
              i = 0,
              len = identifiers.length;

          for (i; i < len; i++) {
            name = waiter.getModuleName(identifiers[i]);
            context[name] = dependencies[i];
          }
        });
      }
      else{
        return $.Deferred().resolve().promise();
      }
    },

    /**
     * Helper method to inspect the state of a module that may or may not be loading or already loaded. Returns undefined
     * if the passed key is not registered.
     *
     * @method  status
     * @param  {String} key Name of desired module
     * @return {String|undefined} Element of the set ["resolved", "pending", "rejected", undefined]
     * @api public
     */
    status: function(key) {
      var module = _store.remoteStack[key] || _store.loadStack[key];

      if (module) {
        return module.state();
      }
    },

    /**
     * Uses settings.timeout to determine how long to wait before calling settings.errorHandler
     *
     * @method  watchForDependencyTimeout
     * @param  {Promise} dfd
     * @param  {String} key
     * @return {void}
     * @api private
     */
    watchForDependencyTimeout: function(dfd, key) {
      var ms = settings.moduleTimeouts[key] || settings.timeout;

      setTimeout(function(){
        if(dfd.state() == 'pending'){
          _store.errorHandler(dfd, {
            data: key,
            msg: 'Module "' + key + '" timed out.'
          });
        }
      }, ms);
    },

    progressHandler: function(index, moduleKey){
      if(settings.debug)
        console.log(moduleKey + ': Try #' + index++);
    },

    errorHandler: function(dfd, envelope){
      dfd.reject(envelope.data);
      if(settings.debug)
        console.log('Error: ' + ( envelope.msg || 'Module failed to load.' ), envelope.data);
    },

    /**
     * Convert a colon-separated namespace `identifier` string to an object
     * name.
     *
     * @method getModuleName
     * @param {String} identifier The colon separated string to convert.
     * @api private
     */
    getModuleName: function(identifier){
      var nameArr = identifier.split(':');
      return (nameArr.length > 1 ? nameArr[1] : '') + (nameArr[0] !== 'Component' ? nameArr[0] : '');
    },

    /**
     * Remove a module from the system.
     *
     * @method unload
     * @param {String} key The key for the module to be removed.
     * @api public
     */
    unload: function(key){
      _store.loadStack[key] = null;
      _store.remoteStack[key] = null;
      if(this[key])
        this[key] = null;
    },

    /**
     * Need to access waiter internals? Wrap your method in a mixin and you'll get access to private methods and data.
     * 
     * @param  {Function} func
     * @return {Function} The wrapped function. The first argument will be an array of arguments, 
     * the second is the private store, the third is an unprotected waiter api.
     * @api public
     */
    mixin: function(func) {
      return function() {
        return func.call(this, arguments, _store, waiter);
      };
    }
  };

  waiter.config();

  window.waiter = {
    load: waiter.load,
    preload: waiter.preload,
    unload: waiter.unload,
    waitFor: waiter.waitFor,
    status: waiter.status,
    mixin: waiter.mixin
  };

})(window, document, jQuery);